/*
 * ARC Linux kernel loader main line -- see _start().
 *
 * All ELF stuff in here.
 * File i/o abstracted to file_open() and file_seek_read().
 *
 * Copyright 1999 Silicon Graphics, Inc.
 */

#include <stddef.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <types.h>

#include <arc.h>
#include <elf.h>

#include <sys/types.h>


#define ANSI_CLEAR	"\033[2J"

#define PAGE_SIZE	4096
#define STACK_PAGES	16


#define CL_MAGIC_ADDR	0x20
#define CL_MAGIC	0xA33F
#define CL_BASE_ADDR	0x90000
#define CL_OFFSET	0x22

#define ALT_MEM_K (*(unsigned long *) (CL_BASE_ADDR + 0x1e0))

#define COMMAND_LINE ((char *) (CL_BASE_ADDR + 2048))
#define COMMAND_LINE_SIZE 256


typedef enum { False = 0, True } Boolean;

/* Difference between virtual and physical addresses */
ULONG voffset = 0xc0000000;

/*
 *  Reserve this memory for loading kernel
 *  Don't put loader structures there because they would be overwritten
 */
ULONG reserve_base = 0x0100000;
ULONG reserve_size = 0x0200000;

CHAR *LoadIdentifier = NULL;
CHAR *OSLoadPartition = NULL;
CHAR *OSLoadFilename = NULL;
CHAR *OSLoadOptions = "";

ULONG mem_top = 0;

static ULONG fd;

/*
 *
 */
static void arc_file_open(const CHAR *partition, const CHAR *filename)
{
	LONG status;
	ULONG nb = strlen(partition) + strlen(filename);
	CHAR *path = (CHAR *) malloc(nb);

	if (path == NULL)
		Fatal("Malloc of %d for %s and %s failed\n",
			nb, partition, filename);

	strcpy(path, partition);
	strcpy(path + strlen(partition), filename + 1); /* skip leading '/' */
	status = ArcOpen(path, OpenReadOnly, &fd);
	if (status != ESUCCESS)
		Fatal("Open of %s failed\n", path);
}

static void arc_seek_read(ULONG off, VOID *cp, ULONG nb, ULONG *got)
{
	LONG status;
	LARGEINTEGER li = { .HighPart = 0, .LowPart = off };

	status = ArcSeek(fd, &li, SeekAbsolute);
	if (status != ESUCCESS)
		Fatal("Seek to 0x%x failed\n", off);

	status = ArcRead(fd, cp, nb, got);
	if (status != ESUCCESS)
		Fatal("Read of 0x%x bytes at 0x%x failed\n", nb, off);
}

void file_open(const char *partition, const char *filename)
{
	return arc_file_open(partition, filename);
}

void file_seek_read(off_t off, char *cp, unsigned int nb, unsigned int *got)
{
	return arc_seek_read((ULONG)off, (VOID*)cp, (ULONG)nb, (ULONG*)got);
}

void InitMalloc(void)
{
	MEMORYDESCRIPTOR *current = NULL;
	ULONG stack = (ULONG) & current;

	current = ArcGetMemoryDescriptor(current);
	while (current != NULL) {
		/*
		 *  The spec says we should have an adjacent FreeContiguous
		 *  memory area that includes our stack.  It would be much
		 *  easier to just look for that and give it to malloc, but
		 *  the 320 only shows FreeMemory areas, no FreeContiguous.
		 *  Oh well.
		 */
		if (current->Type == FreeMemory) {
			ULONG start = current->BasePage * PAGE_SIZE;
			ULONG end = start + (current->PageCount * PAGE_SIZE);

			/* Leave some space for our stack */
			if ((stack >= start) && (stack < end))
				end = (stack - (STACK_PAGES * PAGE_SIZE)) & ~(PAGE_SIZE - 1);

			/* Don't use memory from reserved region */
			if ((start >= reserve_base) && (start < (reserve_base + reserve_size)))
				start = reserve_base + reserve_size;
			if ((end > reserve_base) && (end <= (reserve_base + reserve_size)))
				end = reserve_base;

			if (end > start)
				arclib_malloc_add(start, end - start);
		}

		if (current->BasePage + current->PageCount > mem_top)
			mem_top = current->BasePage + current->PageCount;

		current = ArcGetMemoryDescriptor(current);
	}

	mem_top = mem_top ? mem_top / 256 : 128;
}

void ProcessArguments(LONG argc, CHAR * argv[])
{
	LONG arg;
	CHAR *equals;
	size_t len;

	for (arg = 1; arg < argc; arg++) {
		equals = strchr(argv[arg], '=');
		if (equals != NULL) {
			len = equals - argv[arg];
			if (strncmp(argv[arg], "LoadIdentifier", len) == 0)
				LoadIdentifier = equals + 1;
			if (strncmp(argv[arg], "OSLoadPartition", len) == 0)
				OSLoadPartition = equals + 1;
			if (strncmp(argv[arg], "OSLoadFilename", len) == 0)
				OSLoadFilename = equals + 1;
			if (strncmp(argv[arg], "OSLoadOptions", len) == 0) {
				/* Copy options to local memory to avoid overwrite later */
				OSLoadOptions = strdup(equals + 1);
			}
		}
	}

	/* Workaround for annoying OSLoadPartion behavior in GUI */
	if ((OSLoadFilename != NULL) && (*OSLoadFilename != '/')) {
		CHAR *path = strchr(OSLoadFilename, '/');

		if (path != NULL) {
			OSLoadPartition = malloc((path - OSLoadFilename) + 1);

			if (OSLoadPartition == NULL)
				Fatal ("Cannot allocate memory for argument processing\n");

			strncpy(OSLoadPartition, OSLoadFilename,
				path - OSLoadFilename);
			OSLoadPartition[path - OSLoadFilename] = '\0';
		}

		OSLoadFilename = path;
	}
}

void LoadProgramSegments(Elf32_Ehdr * header)
{
	int idx;
	Boolean loaded = False;
	Elf32_Phdr *segment, *segments;
	size_t size = header->e_phentsize * header->e_phnum;

	if (size <= 0)
		Fatal("No program segments\n");

	segments = malloc(size);
	if (segments == NULL)
		Fatal("Cannot allocate memory for segment headers\n");

	file_seek_read(header->e_phoff, (VOID *)segments, size, NULL);

	segment = segments;
	for (idx = 0; idx < header->e_phnum; idx++) {
		void *memp;
		if (segment->p_type == PT_LOAD) {
			printf
			    ("Loading program segment %u at 0x%x, size = 0x%x\n",
			     idx + 1, segment->p_vaddr - voffset,
			     segment->p_filesz);

			memp = (void *) (segment->p_vaddr - voffset);
			file_seek_read(segment->p_offset, memp,
				       segment->p_filesz, NULL);

			size = segment->p_memsz - segment->p_filesz;
			if (size > 0) {
				printf("Zeroing memory at 0x%x, size = 0x%x\n",
				       (segment->p_vaddr + segment->p_filesz) -
				       voffset, size);
				memset((void
					*) ((segment->p_vaddr +
					     segment->p_filesz) - voffset), 0,
				       size);
			}

			loaded = True;
		}

		segment =
		    (Elf32_Phdr *) (((char *) segment) + header->e_phentsize);
	}

	if (!loaded)
		Fatal("No loadable program segments found\n");

	free(segments);
}

void LoadKernel(const char *partition, const char *filename)
{
	Elf32_Ehdr header;
	extern void StartKernel(ULONG entry, const char *cmdline);
	ULONG entry;

	file_open(partition, filename);

	file_seek_read(0, (void *) &header, sizeof (header), NULL);

	if (memcmp(&(header.e_ident[EI_MAG0]), ELFMAG, SELFMAG) != 0)
		Fatal("Not an ELF file\n");

	if (header.e_ident[EI_CLASS] != ELFCLASS32)
		Fatal("Not a 32-bit file\n");

	if (header.e_ident[EI_DATA] != ELFDATA2LSB)
		Fatal("Not a little-endian file\n");

	if (header.e_ident[EI_VERSION] != EV_CURRENT)
		Fatal("Wrong ELF version\n");

	if (header.e_type != ET_EXEC)
		Fatal("Not an executable file\n");

	if (header.e_machine != EM_386)
		Fatal("Unsupported machine type\n");

	if (header.e_version != EV_CURRENT)
		Fatal("Wrong ELF version\n");

	LoadProgramSegments(&header);

//	entry = (char *) (((ULONG)header.e_entry) - voffset);
	entry = 0x100000;

	printf("Starting kernel; entry point = 0x%x\n", entry
	       /*((ULONG) header.e_entry) - voffset*/);

	asm volatile (
		"cli\n"		/* Disable interrupts */
		"jmp *%0"	/* Here we go! */
		: : "a" (entry), "b" (0), "S" (CL_BASE_ADDR)
	);
}

void setup_boot_parameters(ULONG mem_top, char *command_line)
{
	memset((void *) CL_BASE_ADDR, 0, 4096);

	ALT_MEM_K = mem_top;

	*(USHORT *)(CL_BASE_ADDR + CL_MAGIC_ADDR) = CL_MAGIC;
	*(USHORT *)(CL_BASE_ADDR + CL_OFFSET) = 0x800;

	strncpy(COMMAND_LINE, command_line, COMMAND_LINE_SIZE);
}


void _start(LONG argc, CHAR * argv[], CHAR * envp[])
{
	MEMORYDESCRIPTOR *mem = ArcGetMemoryDescriptor(NULL);

	/* Print identification */
	printf(ANSI_CLEAR "ARC Linux loader 200G2\n\n");

	InitMalloc();

	printf("Memory avaialble: %u MB\n", mem_top);

	ProcessArguments(argc, argv);

	if (LoadIdentifier != NULL)
		printf("LoadIdentifier  :  %s\n", LoadIdentifier);
	if (OSLoadPartition != NULL)
		printf("OSLoadPartition :  %s\n", OSLoadPartition);
	if (OSLoadFilename != NULL)
		printf("OSLoadFilename  :  %s\n", OSLoadFilename);
	if (OSLoadOptions != NULL)
		printf("OSLoadOptions   :  %s\n", OSLoadOptions);
	printf("\n\n");

	if (OSLoadPartition == NULL)
		Fatal("Invalid load parition\n");
	if (OSLoadFilename == NULL)
		Fatal("Invalid load filename\n");

	setup_boot_parameters(mem_top * 1024 * 1024, OSLoadOptions);

	LoadKernel(OSLoadPartition, OSLoadFilename);

	/* Not likely to get back here in a functional state, but what the heck */
	Wait("\n--- Press <spacebar> to restart ---");
	ArcRestart();
}
